/*
 * OpenEAS Experlog Application Server
 * Web application development and dynamic template scripting engine
 * Copyright (c) 2004 ExperLog SARL - http://www.experlog.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version. This program is distributed in
 * the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
 * the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have
 * received a copy of the GNU General Public License along with this program;
 * if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package org.experlog.lite;

import org.experlog.base.OpenClient;
import org.experlog.base.ESStore;
import org.experlog.util.ESCategory;
import org.experlog.util.AppConfiguration;
import org.experlog.util.ESHttpUtils;
import org.apache.log4j.Category;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.RandomAccessFile;

/**
 * Works in conjunction with a browser's form-based file upload
 * capability to allow the client to transfer a file (binary or ASCII)
 * to the server and store it in the server's file system. It requires
 * that the client web browser supports the form-based upload
 * specification documented in RFC 1867, such as Netscape Navigator,
 * MS Internet Explorer and Opera.
 *
 * Parameters:
 * - urles: redirection URL (after upload)
 * - uploaddirectory: directory where files should be upload (relative to
 *   app.UploadDir property)
 * - uploadfile [ uploadfile2 ... uploadfileN ]: file(s) to upload
 * - nbfiles: number of files to upload (default 1)
 * - overwrite: on or off (default on, overwrite existing copies)
 */
public class ESUploaderServlet extends HttpServlet {

	private static final long serialVersionUID = 7154444616162553785L;

	/**
	 * return url when ok
	 */
	private static String urlES = null;

	final private static String fParNbfiles_ = "nbfiles";
	private int mValNbfiles_ = 1;

	/**
	 * Specifies maximum permissible size of file to be uploaded.
	 * Should be passed by calling html document.
	 * <ul><li>parameter name = "maxsize"
	 *     <li>default value  = 1024000 (ie bytes)
	 *     <li>Corresponding instance variable used for containing
	 *         parameter value:<br>{@link #mValMaxsize int mValMaxsize}.
	 * </ul>
	 */
	final private static String fParMaxsize = "maxsize";

	/**
	 * States whether to overwrite any existing version of file
	 * to be uploaded. Should be passed by calling html document.
	 * <ul><li>parameter name        = "overwrite"
	 *     <li>default String value  = "off"
	 *     <li>default boolean value = false
	 *     <li>Corresponding instance variables used for containing
	 *         parameter value:<br>{@link #mValOverwrite String mValOverwrite}
	 *         <br>{@link #mOverwrite boolean mOverwrite}.
	 * </ul>
	 */
	final private static String fParOverwrite = "overwrite";

	/**
	 * Specifies target window where to send result.
	 * Should be passed by calling html document.
	 * <ul><li>parameter name = "uploaddirectory"
	 *     <li>default value  = "" (ie can be empty. Looks for html documents
	 *                              in current directory)
	 *     <li>Corresponding instance variable used for containing
	 *         parameter value:<br>{@link #mValUploaddirectory String mValUploaddirectory}.
	 * </ul>
	 */
	final private static String fParUploaddirectory = "uploaddirectory";

	/**
	 * Specifies target window where to send result.
	 * Should be passed by calling html document.
	 * <ul><li>parameter name = "uploadfile"
	 *     <li>default value  = "" (ie needs to be specified by html document,
	 *                              otherwise a ServletException is thrown)
	 *     <li>Corresponding instance variable used for containing
	 *         parameter value:<br>
	 * </ul>
	 */
	final private static String fParUploadfile = "uploadfile";

	/**
	 * States whether to verify upload. If true checks existance
	 * as well as size of uploaded file. At present file content
	 * is not checked.
	 * <ul><li>parameter name        = "verify"
	 *     <li>default String value  = "off"
	 *     <li>default boolean value = false
	 *     <li>Corresponding instance variables used for containing
	 *         parameter value:<br>{@link #mValVerify String mValVerify}
	 *         <br>{@link #mVerify boolean mVerify}.
	 * </ul>
	 */
	final private static String fParVerify = "verify";

	/** Stores value associated with {@link #fParUploaddirectory} */
	private String mValUploaddirectory = "";
	/** Stores String value associated with {@link #fParOverwrite} */
	private String mValOverwrite = "off";
	/** Stores boolean value associated with {@link #fParOverwrite} */
	private boolean mOverwrite = false;
	/** Stores value associated with {@link #fParUploadfile} */
	private String mValUploadfiles[] = { "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "" };
	/** Stores String value associated with {@link #fParVerify} */
	private String mValVerify = "off";
	/** Stores boolean value associated with {@link #fParVerify} */
	private boolean mVerify = false;
	/** Stores value associated with {@link #fParMaxsize} */
	private int mValMaxsize = 1024000;


	private OpenClient session_;
	private Category logger;

	private void setNoStore(HttpServletResponse res) {
		res.setHeader("Pragma", "no-cache"); // HTTP 1.0
		res.setHeader("Cache-Control", "no-store"); // HTTP 1.1
		res.setDateHeader("Expires", 0); //prevents caching at the proxy server
	}

	public void init(ServletConfig cfg) throws ServletException {
		try {
			super.init(cfg);
			ESStore.getInstance();
		} catch (Exception e) {
			logger = ESCategory.getCategory("openeas.uploadservlet");
			logger.error("ESUploaderServlet:init Exception " + e);

			throw new ServletException(e.getMessage());
		}
	} // method init

	public void doGet(HttpServletRequest req, HttpServletResponse res)
			throws ServletException, IOException {

		try {

			String ct = req.getContentType();

			if (ct == null || ct.indexOf("multipart") == -1) {
				System.err.println("ESUploaderServlet:service UploadServlet request not multipart.");
			} else {
				MultipartRequest mreq = new MultipartRequest(req);
                setNoStore(res); // No cache

                String sessionid = ESHttpUtils.getSessionId(req);
                session_ = ESStore.getInstance().getClient(sessionid, req);

				PrintWriter out = res.getWriter();

				if (!session_.noComment()) {
					out.println("<!-- Generated by openeas http://www.experlog.com -->");
				}

				loadServletParams(mreq, res);
				doUpload(mreq, res);
			}

		} catch (Exception e) {
			if (session_ != null) {
				Category logger = ESCategory.getCategory(ESStore.getInstance().getShopName(), "openeas.lite");
				logger.error("ERROR IN UploaderServlet " + e);
			}
			e.printStackTrace(System.err);
			throw new ServletException(e.getMessage());
		}
	}

	public void doPost(HttpServletRequest req, HttpServletResponse res)
			throws ServletException, IOException {
		doGet(req, res);
	}

	public String getServletInfo() {
		return "ESUploader servlet";
	}

	/**
	 * Loads servlet parameters from incoming request required for uploading.
	 * @param mreq MultipartRequest containing data required for uploading client data
	 * @param res  HttpServletResponse containing servlet (ie server) response to client
	 * @throws ServletException if any of the required parameter missing/invalid. Required
	 *                          parameter/values:
	 *                          <ul><li>content type should be multipart/form-data
	 *                              <li>name of file to be uploaded should contain
	 *                                  a non-empty value</ul>
	 */
	protected void loadServletParams(MultipartRequest mreq, HttpServletResponse res) throws ServletException {
		System.err.println("ESUploaderServlet:loadServletParams Starting UploadServlet ... ");
		String max = getParameter(mreq, fParMaxsize, "" + mValMaxsize);
		urlES = getParameter(mreq, "urles", "urles");

		try {
			mValMaxsize = Integer.parseInt(max);
		} catch (NumberFormatException ignore) {
		}
		String ctype = mreq.getContentType();

		//make sure content type is multipart/form-data
		if (ctype == null || ctype.indexOf("multipart/form-data") == -1) {
			throwServletException(mreq, "Error while accessing content type parameter."
										+ "\nError description: parameter missing or malformed.");
		}

		String nbf = getParameter(mreq, fParNbfiles_, "1");
		try {
			mValNbfiles_ = Integer.parseInt(nbf);
		} catch (NumberFormatException nfe) {
			mValNbfiles_ = 1;
		}

		for(int i=1; i<=mValNbfiles_; i++) {

			String paramName = fParUploadfile + (i > 1 ? Integer.toString(i) : "");

			String sourceFile = mreq.getFileName(paramName);
			if(sourceFile == null)
				throw new ServletException("Missing parameter: " + paramName);

			if(!AppConfiguration.isSecurePath(sourceFile))
				throw new ServletException(
					"Access forbidden for security reasons "
					+ "(tried to access parent directory)");

			mValUploadfiles[i-1] = mreq.getShortFileName(paramName);
		}

		mValUploaddirectory = getParameter(mreq, fParUploaddirectory, "");

		if (!AppConfiguration.isSecurePath(mValUploaddirectory)) {
			throw new ServletException(
					"Access forbidden for security reasons "
					+ "(tried to access parent directory)");
		}

		mOverwrite = getParameter(mreq, fParOverwrite, mValOverwrite).toLowerCase().equals("on");
		mVerify = getParameter(mreq, fParVerify, mValVerify).toLowerCase().equals("on");
		res.setContentType("text/html");
		System.err.println("ESUploaderServlet:loadServletParams UploadServlet done. ");
	}

	/**
	 * Reads the uploaded data from the request and writes it to a file.
	 * @param mreq MultipartRequest containing client data to be uploaded
	 * @param res  HttpServletResponse containing servlet (ie server) response to client
	 * @throws ServletException if uploading fails (eg because file already exists
	 *                          and mOverwrite set to false)
	 */
	protected void doUpload(MultipartRequest mreq, HttpServletResponse res) throws ServletException {

		String upld = session_.getAppConfiguration().getUploadDir();
		if (upld == null) {
			throw new ServletException("The app.UploadDir property is not present in the configuration file (session=" + session_ + ")");
		}
		String fullDir = buildFileName(upld, mValUploaddirectory);

		System.err.println("ESUploaderServlet:doUpload UploadServlet doUpload ... dir=" + fullDir);

		//create File object to check for existence of Directory
		File fileDir = new File(fullDir);
		if (!fileDir.exists()) {
			//Directory doesn't exist, create it
			fileDir.mkdirs();
		}

        String parameters = "uploaddir="+mValUploaddirectory;
        int nbfilez = 0;
		for(int i=1; i<=mValNbfiles_; i++) {

			// Ignore empty file names
			if(mValUploadfiles[i-1].trim().length() < 1) continue;
            nbfilez ++;

			String fullFileName = buildFileName(fullDir, mValUploadfiles[i-1]);
			File fileUpload = new File(fullFileName);
			if (fileUpload.exists() && !mOverwrite) {
				throwServletException(mreq, "File exists already.");
			}

			//instantiate file output stream
			FileOutputStream outUpload = null;
			try {
				//System.err.println("fullFileName = " + fullFileName);
				outUpload = new FileOutputStream(fullFileName);
			} catch (FileNotFoundException e) {
				throwServletException(mreq, "Error while opening file: " + e);
			}

			String paramName = fParUploadfile + (i > 1 ? Integer.toString(i) : "");
			byte data[] = mreq.getFileValue(paramName);
			if(data == null) { // null if I try to upload a non-existent file
				data = ("Error: file " + fullFileName + " does not exist.").getBytes();
			}
			try {
				outUpload.write(data, 0, data.length);
				outUpload.close();
			} catch (Exception ex) {
				ex.printStackTrace(System.err);
				throwServletException(mreq, "Error while uploading file." + ex);
			}
            //System.err.println("End of the name:"+mreq.getShortFileName(paramName));
            parameters += "&uploadfile"+i+"="+mreq.getShortFileName(paramName);

			if (mVerify) {
				try {
					RandomAccessFile raf = new RandomAccessFile(fullFileName, "r");
					long length = raf.length();
					if (length != data.length) {
						throw new Exception("Verification failed. Incorrect length: " + raf.length());
					}
				} catch (FileNotFoundException fnf) {
					throwServletException(mreq, "\nException: " + fnf);
				} catch (Exception ex) {
					throwServletException(mreq, "\nException: " + ex);
				}
			}
		}
        parameters += "&nbfiles=" + nbfilez;
		try {
            System.err.println("parameters="+parameters);
            if (urlES.indexOf('?') == -1)
                urlES = urlES+"?"+parameters;
            else
                urlES = urlES+"&"+parameters;

            System.err.println("URL="+urlES);
            res.sendRedirect(urlES);
		} catch (IOException e) {
			throwServletException(mreq, "\nException: " + e);
		}

		System.err.println("ESUploaderServlet:doUpload UploadServlet done. ");
	}

	// ========== auxilliary methods ========== //

	/**
	 * Extracts parameter value from incoming request if available.
	 * @param name         String containing parameter name
	 * @param defaultValue String containing default parameter value
	 * @return             String containing requested parameter value.
	 */
	private String getParameter(MultipartRequest mreq, String name, String defaultValue) {
		String value = mreq.getParameter(name);
		if (value == null) {
			value = defaultValue;
		}
		return value;
	}

	/**
	 * Creates a html document for presenting ServletException that is to be thrown
	 * @param mreq    MultipartRequest containing incoming request URL to be passed to
	 * @param message String contianing message to be inserted into ServletException
	 * @throws        ServletException due to error occurring in calling method
	 */
	private void throwServletException(MultipartRequest mreq, String message) throws ServletException {
		String resultHtml = "<HTML><HEAD><TITLE>UploadResult</TITLE></HEAD>"
				+ "<BODY BGCOLOR=\"#ffffff\"><P><FONT color=\"red\"><B>"
				+ message + "<p>" + "</B></BODY></HTML>";
		throw new ServletException(resultHtml);
	}

	/**
	 * Concatenates specified directory and file name values to create fully-qualified
	 * file name.
	 * @param filename  String containing file name to be concatenated to specified
	 *                  directory
	 * @param directory String containing path to be concatenated to specified file
	 *                  name
	 */
	private String buildFileName(String directory, String filename) {
		if (directory == null) {
			directory = "";
		}
		if (filename == null) {
			filename = "";
		}

		directory = directory.trim();
		filename = filename.trim();

		if (! (directory.endsWith("\\") || directory.endsWith("/"))) {
			directory += File.separator;
		}

		if (filename.startsWith("\\") || filename.startsWith("/")) {
			if (filename.length() > 1) {
				filename = filename.substring(1, filename.length());
			} else {
				filename = "";
			}
		}
		return directory + filename;
	}
}

